function [pv,ev,Sg1,Exp1,p_S0]=upper_bound(S,m,u,n,f,r,Ti,Tf)

%   This function uses a modification of an algorithm in the paper
%      "Numerical Computation of Multivariate Normal Probabilities", 
%      A. Genz, J. of Comp. Graph. Stat., 1(1992), pp. 141-149.
%
%   The primary references for the numerical integration are 
%    "Randomization of Number Theoretic Methods for Multiple Integration",
%     R. Cranley & T.N.L. Patterson, SIAM J Numer Anal, 13(1976), pp. 904-14.
%   and  
%    "Fast Component-by-Component Construction, a Reprise for Different 
%     Kernels", D. Nuyens & R. Cools. In H. Niederreiter and D. Talay,
%     editors, Monte-Carlo and Quasi-Monte Carlo Methods 2004, 
%     Springer-Verlag, 2006, 371-385.


eps1 = 5e-9; 
pv = zeros(1,n);
ev = zeros(1,n);
df = zeros(S,n);
[Sg1,Sg2,Exp1,Exp2,r_2]=Cholesky(f,r,n,Ti,Tf);
a = Sg1(1,1);b = r_2(1); c = Sg2(1,1);d=a*b-c^2;
p_S0 = phi((u-Exp1(1))/sqrt(a))*(phi((a*Exp2(1) + c*(u-Exp1(1)))/sqrt(a*d))/a + (a*Exp2(1) + c*(u-Exp1(1)))/a/sqrt(d*a)*(1-Phi2((-(a*Exp2(1) + c*(u-Exp1(1))/sqrt(a*d))))));

for i=1:n,
    Sg = zeros(i+2);
        Sg(1,1)=Sg1(i+1,i+1); 
        Sg(2,1)=Sg2(i+1,i+1); Sg(1,2) = Sg(2,1); Sg(2,2)=r_2(i+1); 
        Sg(3:i+2,1) = Sg1(1:i,i+1); Sg(1,3:i+2) = Sg(3:i+2,1)';
        Sg(2,3:i+2)=Sg2(i+1,1:i); Sg(3:i+2,2) = Sg(2,3:i+2)';
        Sg(3:i+2,3:i+2) = Sg1(1:i,1:i); %Covariance matrix
        
        [ U, D ] = svd(Sg);
        d = sqrt(diag(D));
        nc =i+2;
        if d(nc) < eps1, 
            nc = min( find( d < eps1 ) );
        end
        [ Q, R ] = qr( diag(d(1:nc))*U(:,1:nc)' );
        ch = ( diag(sign(diag(R)))*R )';  %Cholesky for S_i
       
        if nc > 64, q = sqrt(primes(5*nc*log(nc)/4))'; %Richtmyer generators
        else [q pr] = fstrnk(m,nc-2);  %Lattice Rule 
            q = q/pr;
        end
        
        uv = (u - Exp1(i+1))/ch(1,1)*ones(1,m);
        N_1 = min(max(Phi2((-Exp2(i+1)-ch(2,1)*(u-Exp1(i+1))/ch(1,1))/ch(2,2)),-1e30),1e30);
        pu = phi(uv)*(1-N_1);
        
        for K = 1:S,
            fi = zeros(1,m); y = zeros(nc-2,m);
            y(1,:) = Phinv(abs(2*mod(q(1)*[1:m]+rand,1) - 1 )*(1-N_1) + N_1);
            fi = pu.*(ch(2,2)*y(1,:)+ ch(2,1)*uv + Exp2(i+1)*ones(1,m))/ch(1,1);
            
                    bi = 9;
                    ai = -9;
        
                for j = 2 : nc-2, 
                    ys = ch(j+1,1:j)*[uv;y(1:j-1,:)];
                    dc = Phi2( min( ( u - Exp1(j-1) - ys )/ch(j+1,j+1), 9 ) );
                    fi = fi.*dc;
                    y(j,:) = Phinv(abs( 2*mod(q(j)*[1:m]+rand,1) - 1 ).*dc ); % MCQMC
                end
               
                 for j = nc:i+2,
                     
                    ys = ch(j,1:nc-1)*[ uv; y(1:nc-2,:) ];
                      if ch(j,nc) > 0,
                           bi = min( ( u - Exp1(j-2) - ys )/ch(j,nc), bi );
                           else
                           ai = max( ( u - Exp1(j-2) - ys )/ch(j,nc), ai );
                      end
                 end
                
            df(K,i) = mean(fi.*(Phi2(bi) - Phi2(ai))); %Estimation of S_i 
        end
end
pv = mean(df);
ev = 1/(S*(S-1))*sum((df - ones(S,1)*pv).^2); ev = 3*sqrt(ev);
end
